# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2021, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.

from biorefineries.cane.composition import (
    set_lipid_fraction as set_oil_fraction,
    get_composition,
)
from biorefineries.oilcane._load_data import load_composition as load_feedstock_composition
from biorefineries import oilcane as oc
import chaospy as cp

__all__ = (
    'OilExtractionSpecification', 
    'MockExtractionSpecification',
    'CaneCompositionSpecification',
    'set_composition_parameters',
    'set_line_composition_parameters',
)

class MockExtractionSpecification:
    
    def load_saccharification_oil_recovery(self, bagasse_recovery):
        pass
      
    def load_crushing_mill_oil_recovery(self, recovery):
        pass
    
    def load_specifications(self, 
            recovery=None,
            bagasse_recovery=None,\
        ):
        pass

class OilExtractionSpecification:
    """
    Create a OilExtractionSpecification object for setting process 
    specifications related to oil extraction.
    
    Parameters
    ----------
    system : System
        System associated to feedstocks.
    isplit_crushing_mill : ChemicalIndexer
        Defines extraction recovery as a material split.
        
    Notes
    -----
    Use the `load_specifications` method to load specifications.
    
    """
    
    __slots__ = (
        'feedstocks',
        'isplit_crushing_mill',
        'isplit_saccharification',
        'crushing_mill_oil_recovery',
        'saccharification_oil_recovery',
        'isplit_recovery_is_reversed',
    )
    
    def __init__(self, system, isplit_crushing_mill, isplit_saccharification=None):
        self.isplit_crushing_mill = isplit_crushing_mill #: [SplitIndexer] Defines extraction recovery as a material split.
        self.isplit_saccharification = isplit_saccharification #: [SplitIndexer] Defines extraction recovery from bagasse as a material split.
        if isplit_crushing_mill is not None: 
            self.crushing_mill_oil_recovery = isplit_crushing_mill['Oil'].mean()
        else:
            self.crushing_mill_oil_recovery = None
        if isplit_saccharification is not None: 
            self.saccharification_oil_recovery = 1 - isplit_saccharification['Oil'].mean()
        else:
            self.saccharification_oil_recovery = None
      
    def load_crushing_mill_oil_recovery(self, recovery):
        if self.isplit_crushing_mill is None: return
        self.crushing_mill_oil_recovery = recovery
        self.isplit_crushing_mill['Oil'] = 1. - recovery
    
    def load_saccharification_oil_recovery(self, recovery):
        if self.isplit_saccharification is None: return
        self.saccharification_oil_recovery = recovery
        self.isplit_saccharification['Oil'] = 1. - recovery
    
    def load_specifications(self, 
            crushing_mill_oil_recovery=None,
            saccharification_oil_recovery=None,
        ):
        """
        Load oil extraction specifications.

        Parameters
        ----------
        recovery : float, optional
            Oil extraction recovery.
        bagasse_recovery : float, optional
            Oil extraction recovery from bagasse.
        oil_content : float, optional
            Oil content of feedstocks [dry wt. %].

        """
        if crushing_mill_oil_recovery is None: 
            crushing_mill_oil_recovery = self.crushing_mill_oil_recovery
        if saccharification_oil_recovery is None:
            saccharification_oil_recovery = self.saccharification_oil_recovery
        self.load_crushing_mill_oil_recovery(crushing_mill_oil_recovery)
        self.load_saccharification_oil_recovery(saccharification_oil_recovery)
    
    
class CaneCompositionSpecification:
    """
    Create a CaneCompositionSpecification object for setting the 
    composition of cane feedstocks.
    
    Parameters
    ----------
    feedstock : Stream
        Cane feedstock.
    oil : float 
        Oil content of feedstocks [dry wt. %].
    moisture : float 
        Moisture content of feedstocks [wt. %].
    fiber : float 
        Fiber content of feedstocks [dry wt. %].
    FFA : float 
        FFA content of feedstocks [oil wt. %].
    PL : float 
        PL content of feedstocks [oil wt. %].
    
    """
    __slots__ = (
        'feedstock',
        'oil',
        'FFA',
        'PL',
        'moisture',
        'fiber',
    )
    
    def __init__(self, feedstock, oil=None, moisture=None, fiber=None, FFA=None, PL=None):
        composition = get_composition(feedstock) # Get default values
        self.feedstock = feedstock #: [Stream] Cane feedstock.
        self.oil = composition['lipid'] if oil is None else oil #: [float] Oil content of feedstocks [dry wt. %].
        self.moisture = composition['moisture'] if moisture is None else moisture
        self.fiber = composition['fiber'] if fiber is None else fiber
        self.FFA = composition['FFA'] if FFA is None else FFA
        self.PL = composition['PL'] if PL is None else PL
        
    def load_oil_content(self, oil=None, **kwargs):
        """
        Set the oil content [dry wt. %] of cane feedstocks assuming that an 
        increase in oil content is accompanied by a decrease in sugar content 
        using an energy balance and an increase in fiber using a mass balance 
        (oil is more energy dense than sugar).
        
        """
        if oil is None: oil = self.oil
        set_oil_fraction(
            oil, self.feedstock, PL_fraction=self.PL, 
            FFA_fraction=self.FFA, **kwargs,
        )
        self.oil = oil
        
    def load_composition(self, oil=None, moisture=None, fiber=None, FFA=None, PL=None):
        if oil is None: oil = self.oil
        if moisture is None: moisture = self.moisture
        if fiber is None: fiber = self.fiber
        if FFA is None: FFA = self.FFA
        if PL is None: PL = self.PL
        load_feedstock_composition(self.feedstock, oil, moisture, fiber, FFA, PL)
        try:
            assert (self.feedstock.mol >= 0.).all()
        except:
            breakpoint()
    
    
def bounded_gaussian_distribution_from_mean_and_std(mean, std):
    if mean == 0: mean = 1e-6
    if std == 0: std = mean / 1e3
    normal = cp.Normal(mean, std)
    trunc = cp.Trunc(normal, lower=mean - 2 * std, upper=mean + 2 * std)
    return trunc

def set_line_composition_parameters(line):
    df = oc.get_composition_data()
    line = df.loc[line]
    oil = line['Stem Oil (dw)']
    water = line['Water (wt)']
    fiber = line['Fiber (dw)']
    biomass = line['Dry biomass yield (MT/hc)']
    set_composition_parameters(
        oil['Mean'],
        oil['STD'],
        fiber['Mean'],
        fiber['STD'],
        water['Mean'],
        water['STD'],
        biomass['Mean'],
        biomass['STD'],
    )
    
def set_composition_parameters(
        mean_oil_content,
        std_oil_content,
        mean_fiber_content,
        std_fiber_content,
        mean_moisture_content,
        std_moisture_content,
        mean_biomass_yield,
        std_biomass_yield,
    ):
    try:
        _reset_composition_parameters_for_cane_line(
            mean_oil_content,
            std_oil_content,
            mean_fiber_content,
            std_fiber_content,
            mean_moisture_content,
            std_moisture_content,
            mean_biomass_yield,
            std_biomass_yield,
        )
    except:
        _add_model_composition_parameters_for_cane_line(
            mean_oil_content,
            std_oil_content,
            mean_fiber_content,
            std_fiber_content,
            mean_moisture_content,
            std_moisture_content,
            mean_biomass_yield,
            std_biomass_yield,
        )

def _add_model_composition_parameters_for_cane_line(
        mean_oil_content,
        std_oil_content,
        mean_fiber_content,
        std_fiber_content,
        mean_moisture_content,
        std_moisture_content,
        mean_biomass_yield,
        std_biomass_yield,
        model=None,
    ):
    if model is None: model = oc.model
    removed = {oc.set_cane_oil_content, oc.set_relative_sorghum_oil_content}
    model.parameters = tuple([i for i in model.parameters if i not in removed])
    parameter = model.parameter
    @parameter(
        element=oc.feedstock, 
        baseline=mean_oil_content,
        distribution=bounded_gaussian_distribution_from_mean_and_std(
            mean_oil_content, std_oil_content
        ),
        units='dw %',
    )
    def set_cane_oil_content(oil_content):
        oc.composition_specification.oil = oil_content
    
    @parameter(
        element=oc.feedstock, 
        baseline=mean_fiber_content,
        distribution=bounded_gaussian_distribution_from_mean_and_std(
            mean_fiber_content, std_fiber_content
        ),
        units='dw %',
    )
    def set_cane_fiber_content(fiber_content):
        oc.composition_specification.fiber = fiber_content
    
    @parameter(
        element=oc.feedstock, 
        baseline=mean_moisture_content,
        distribution=bounded_gaussian_distribution_from_mean_and_std(
            mean_moisture_content, std_moisture_content
        ),
        units='wt %',
    )
    def set_cane_moisture_content(moisture_content):
        oc.composition_specification.moisture = moisture_content
        oc.composition_specification.load_composition()
        
    @parameter(
        element=oc.feedstock, 
        baseline=mean_biomass_yield,
        distribution=bounded_gaussian_distribution_from_mean_and_std(
            mean_biomass_yield, std_biomass_yield
        ),
        units='dry MT/hc',
    )
    def set_cane_biomass_yield(biomass_yield):
        oc.dry_biomass_yield = biomass_yield
        
    oc.set_cane_oil_content = set_cane_oil_content
    oc.set_cane_fiber_content = set_cane_fiber_content
    oc.set_cane_moisture_content = set_cane_moisture_content
    oc.set_cane_biomass_yield = set_cane_biomass_yield
    
def _reset_composition_parameters_for_cane_line(
        mean_oil_content,
        std_oil_content,
        mean_fiber_content,
        std_fiber_content,
        mean_moisture_content,
        std_moisture_content,
        mean_biomass_yield,
        std_biomass_yield,
        parameters=None, 
    ):
    if parameters is None: 
        parameters = (
            oc.set_cane_oil_content, 
            oc.set_cane_fiber_content, 
            oc.set_cane_moisture_content,
            oc.set_cane_biomass_yield,
        )
    oil, fiber, moisture, biomass = parameters
    oil.baseline = mean_oil_content
    oil.distribution = bounded_gaussian_distribution_from_mean_and_std(
        mean_oil_content, std_oil_content
    )
    fiber.baseline = mean_fiber_content
    fiber.distribution = bounded_gaussian_distribution_from_mean_and_std(
        mean_fiber_content, std_fiber_content
    )
    moisture.baseline = mean_moisture_content
    moisture.distribution = bounded_gaussian_distribution_from_mean_and_std(
        mean_moisture_content, std_moisture_content
    )
    biomass.baseline = mean_biomass_yield
    biomass.distribution = bounded_gaussian_distribution_from_mean_and_std(
        mean_biomass_yield, std_biomass_yield
    )
    


# -*- coding: utf-8 -*-
"""
.. autofunction:: biorefineries.cane.systems.sugarcane.create_sugarcane_to_ethanol_combined_1_and_2g
.. autofunction:: biorefineries.cane.systems.sugarcane.create_sucrose_to_ethanol_system
.. autofunction:: biorefineries.cane.systems.sugarcane.create_sugarcane_to_sugar_and_molasses_system
.. autofunction:: biorefineries.cane.systems.sugarcane.create_sugarcane_to_sugar_and_ethanol_system

"""
from . import ethanol
from . import sugar_ethanol
from . import cellulosic_ethanol 

__all__ = (
    *ethanol.__all__,
    *sugar_ethanol.__all__,
    *cellulosic_ethanol.__all__,
)

from .ethanol import *
from .sugar_ethanol import *
from .cellulosic_ethanol import *




# -*- coding: utf-8 -*-
"""
.. contents:: :local:
    
.. autoclass:: biorefineries.biodiesel.units.GlycerolysisReactor
.. autoclass:: biorefineries.biodiesel.units.BlendingTankWithSkimming
.. autoclass:: biorefineries.biodiesel.units.Transesterification

"""
from . import glycerolysis_reactor
from . import skimming_tank
from . import transesterification

from .glycerolysis_reactor import *
from .skimming_tank import *
from .transesterification import *

__all__ = (
    *glycerolysis_reactor.__all__, 
    *skimming_tank.__all__,
    *transesterification.__all__,
)


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
import biosteam as bst
from biosteam import main_flowsheet as f
from biosteam import SystemFactory
from biorefineries.biodiesel import (
    create_lipid_pretreatment_system as create_oil_pretreatment_system,
    create_transesterification_and_biodiesel_separation_system,
)
from ..juicing import (
    create_juicing_system,
    create_feedstock_handling_system,
)
from ..sugarcane import create_sucrose_to_ethanol_system
from ..lipid_extraction import create_post_fermentation_oil_separation_system
from ... import streams as s

__all__ = (
    'create_oilcane_to_biodiesel_and_ethanol_1g',
)

@SystemFactory(
    ID='oilcane_sys',
    ins=[s.oilcane],
    outs=[s.ethanol, s.biodiesel, s.crude_glycerol, s.vinasse], 
)
def create_oilcane_to_biodiesel_and_ethanol_1g(
        ins, outs,
        evaporator_and_beer_column_heat_integration=True,
        WWT_kwargs=None,
    ):
    oilcane, = ins
    ethanol, biodiesel, crude_glycerol, vinasse = outs
    
    feedstock_handling_sys = create_feedstock_handling_system(
        ins=oilcane,
        outs='',
        mockup=True,
        area=100
    )
    
    ### Oil and juice separation ###
    
    juicing_sys, jdct = create_juicing_system(
        ins=feedstock_handling_sys-0,
        outs=['', 'bagasse', 'fiber_fines'],
        pellet_bagasse=False,
        dry_bagasse=True,
        mockup=True,
        udct=True,
        area=200,
    )
    screened_juice, bagasse, fiber_fines = juicing_sys.outs
    # bagasse_pelleting_sys = create_bagasse_pelleting_system(None, bagasse, area=200, mockup=True)
    # pelleted_bagasse, = bagasse_pelleting_sys.outs
    jdct['S201'].isplit['Lipid'] = 1.
    crushing_mill = jdct['U201']
    crushing_mill.tag = "oil extraction"
    crushing_mill.isplit['Lipid'] = 0.90
    
    ### Ethanol section ###
    
    ethanol_production_sys, epdct = create_sucrose_to_ethanol_system(
        ins=[screened_juice, 'denaturant'],
        outs=[ethanol, '', '', ''],
        mockup=True,
        area=300,
        add_urea=True,
        udct=True,
    )
    ethanol, stillage, stripper_bottoms_product, evaporator_condensate_a = ethanol_production_sys.outs
    post_fermentation_oil_separation_sys = create_post_fermentation_oil_separation_system(
        ins=stillage,
        outs=['', '', ''],
        mockup=True,
        area=400,
    )
    oil, thick_vinasse, evaporator_condensate_b = post_fermentation_oil_separation_sys.outs
    oil_pretreatment_sys, oil_pretreatment_dct = create_oil_pretreatment_system(
        ins=oil,
        mockup=True,
        outs=['', 'polar_lipids', ''],
        area=500,
        udct=True,
    )
    MX = bst.Mixer(400, [thick_vinasse, evaporator_condensate_a], vinasse)
    oil, polar_lipids, wastewater = oil_pretreatment_sys.outs
    
    
    ### Biodiesel section ###
    
    # Fresh degummed oil
    transesterification_and_biodiesel_separation_sys, tbdct = create_transesterification_and_biodiesel_separation_system(
        ins=oil, 
        outs=[biodiesel, crude_glycerol, ''],
        mockup=True,
        area=500,
        udct=True,
    )
    MX1 = bst.Mixer(500, [transesterification_and_biodiesel_separation_sys-2, wastewater], 'wastewater')

    ### Facilities ###
    s = f.stream
    u = f.unit
    MX2 = bst.Mixer(600,
        [polar_lipids, bagasse]
    )
    # Burn bagasse from conveyor belt
    bst.CoolingTower(700)
    MX3 = bst.Mixer(700, [evaporator_condensate_b, stripper_bottoms_product], 'recycle_process_water')
    bst.ChilledWaterPackage(700)
    
    if WWT_kwargs:
        outs.remove(vinasse)
        MX1.ins.append(vinasse)
        wastewater_treatment_sys = bst.create_wastewater_treatment_system(
            area=900, ins=[MX1-0], mockup=True, **WWT_kwargs
        )
        treated_water = wastewater_treatment_sys.get_outlet('RO_treated_water')
        sludge = wastewater_treatment_sys.get_outlet('sludge')
        biogas = wastewater_treatment_sys.get_outlet('biogas')
        MX2.ins.append(sludge)
    else:
        treated_water = ''
        biogas = ''
        
    bst.ProcessWaterCenter(700, [treated_water, 'makeup_RO_water', MX3-0, 'makeup_water'], 
                           process_water_price=0.000254)
    
    bst.BoilerTurbogenerator(600,
        (MX2-0, biogas, 
         'boiler_makeup_water',
         'natural_gas',
         'FGD_lime',
         'boilerchems'),
        ('emissions', 'rejected_water_and_blowdown', 'ash_disposal'),
        turbogenerator_efficiency=0.85
    )
    
    def get_hx_issues():
        hxs = [u.E301, u.D501.boiler, u.D502.boiler, u.H501, u.H502, u.H503, u.H504, oil_pretreatment_dct['F3'], u.H506]
        condenser = getattr(u.E401, 'condenser', None)
        if condenser: hxs.append(condenser)
        return hxs
    
    HXN = bst.HeatExchangerNetwork(800, 
        ignored=get_hx_issues,
        Qmin=1e5,
    )
    HXN.acceptable_energy_balance_error = 0.01



# -*- coding: utf-8 -*-
"""
"""
import numpy as np
import biosteam as bst
from warnings import warn
from warnings import filterwarnings
from biorefineries import cane
from chaospy import distributions as shape
from .feature_mockups import (
    all_metric_mockups, 
)
from .results import (
    monte_carlo_file,
    autoload_file_name,
    spearman_file,
)

__all__ = (
    'evaluate_configurations_across_recovery_and_oil_content',
    'evaluate_configurations_across_sorghum_and_cane_oil_content',
    'evaluate_metrics_oil_recovery_integration',
    'evaluate_metrics_at_biomass_yield',
    'run_uncertainty_and_sensitivity',
    'save_pickled_results',
    'run_all',
    'run_sugarcane_microbial_oil_and_ethanol',
    'run_oilcane_microbial_oil_and_ethanol_across_oil_content',
    'run_oilcane_microbial_oil_and_ethanol_across_lines',
)

configuration_names = (
    'S1', 'O1', 'S2', 'O2', 'S1*', 'O1*', 'S2*', 'O2*', 'O3', 'O4', 
)
comparison_names = ( 
    'O1 - S1', 'O2 - S2',  'O2 - O1', 'O1* - O1',  'O2* - O2',
)

def no_derivative(f):
    def f_derivative_disabled(*args, **kwargs):
        cane.Biorefinery.disable_derivative()
        try:
            return f(*args, **kwargs)
        finally:
            cane.Biorefinery.enable_derivative()
    return f_derivative_disabled

def evaluate_configurations_across_recovery_and_oil_content(
        recovery, oil_content, configurations, 
    ):
    A, B = configurations.shape
    data = np.zeros([A, B, N_metrics])
    for index, configuration in np.ndenumerate(configurations):
        br = cane.Biorefinery(configuration)
        if br.configuration.agile:
            br.cane_mode.oil_content = br.sorghum_mode.oil_content = oil_content
            br.oil_extraction_specification.load_crushing_mill_oil_recovery(recovery)
        else:
            br.oil_extraction_specification.load_specifications(
                crushing_mill_oil_recovery=recovery, 
                oil_content=oil_content, 
            )
        try:
            br.sys.simulate()
        except:
            br.sys.empty_recycles()
            br.sys.simulate()
        data[index] = [j() for j in br.model.metrics]
    return data

N_metrics = len(all_metric_mockups)
evaluate_configurations_across_recovery_and_oil_content = no_derivative(
    np.vectorize(
        evaluate_configurations_across_recovery_and_oil_content,
        signature=f'(),(),(a, b)->(a, b, {N_metrics})'
    )
)
def evaluate_configurations_across_sorghum_and_cane_oil_content(
        sorghum_oil_content, cane_oil_content, configurations,
    ):
    C = len(configurations)
    data = np.zeros([C, N_metrics])
    for ic in range(C):
        br = cane.Biorefinery([int(configurations[ic]), True, False])
        br.cane_mode.oil_content = cane_oil_content
        br.sorghum_mode.oil_content = sorghum_oil_content
        br.sys.simulate()
        data[ic, :] = [j() for j in br.model.metrics]
    return data

evaluate_configurations_across_sorghum_and_cane_oil_content = no_derivative(
    np.vectorize(
        evaluate_configurations_across_sorghum_and_cane_oil_content, 
        excluded=['configurations'],
        signature=f'(),(),(c)->(c,{N_metrics})'
    )
)

@no_derivative
def evaluate_metrics_at_composition(oil, fiber):
    data = np.zeros([N_metrics, 2])
    for i, configuration in enumerate(['O2', 'O8']):
        br = cane.Biorefinery(configuration)
        cs = br.composition_specification
        if br.ROI_target is None:
            if configuration == 'O2':
                S2 = cane.Biorefinery('S2')
                S2.set_cane_oil_content.setter(0)
                S2.sys.simulate()
                br.ROI_target = S2.ROI()
            else:
                br.set_cane_oil_content.setter(0)
                br.composition_specification.oil = -1 # Not zero so that competitive biomass metric works
                br.sys.simulate()
                br.ROI_target = br.ROI()
        water = 0.65
        if oil == 0 and configuration == 'O2':
            data[i, :] = [np.nan for i in br.model.metrics]
            continue
        try:
            cane.load_composition(br.feedstock, oil, water, fiber, cs.FFA, cs.PL)
        except ValueError:
            data[i, :] = [np.nan for i in br.model.metrics]
            continue
        br.sys.simulate()
        data[:, i] = [i() for i in br.model.metrics]
    return data

@no_derivative
def evaluate_metrics_at_biomass_yield(oil, dry_biomass_yield):
    data = np.zeros([N_metrics, 2])
    for i, configuration in enumerate(['O7', 'O9']):
        br = cane.Biorefinery(configuration)
        br.composition_specification.load_oil_content(oil / 100, moisture=0.65)
        br.dry_biomass_yield = dry_biomass_yield
        br.update_feedstock()
        br.sys.simulate()
        data[:, i] = [i() for i in br.model.metrics]
    return data

@no_derivative
def evaluate_metrics_oil_recovery_integration(microbial_oil_recovery, microbial_oil_yield, productivity):
    nrows = productivity.size
    data = np.zeros([nrows, 2, N_metrics])
    for i in range(nrows):
        for j, configuration in enumerate(['O9', 'O8']):
            br = cane.Biorefinery(configuration)
            br.set_fermentation_microbial_oil_productivity.setter(productivity[i])
            br.set_microbial_oil_recovery.setter(microbial_oil_recovery)
            br.set_glucose_to_microbial_oil_yield.setter(microbial_oil_yield)
            br.set_xylose_to_microbial_oil_yield.setter(microbial_oil_yield)
            br.sys.simulate()
            data[i, j, :] = [i() for i in br.model.metrics]
    return data

def save_pickled_results(N, configurations=None, rule='L', optimize=True):
    from warnings import filterwarnings
    filterwarnings('ignore', category=bst.exceptions.DesignWarning)
    filterwarnings('ignore', category=bst.exceptions.CostWarning)
    if configurations is None: configurations = configuration_names
    for name in configurations:
        br = cane.Biorefinery(name)
        np.random.seed(1)
        samples = br.model.sample(N, rule)
        br.model.load_samples(samples, optimize=optimize, ss=False)
        file = monte_carlo_file(name, False)
        br.model.load_pickled_results(
            file=autoload_file_name(name),
            safe=False
        )
        br.model.table.to_excel(file)
        br.model.table = br.model.table.dropna(how='all', axis=1)
        for i in br.model.metrics:
            if i.index not in br.model.table: br.model._metrics.remove(i)
        br.model.table = br.model.table.dropna(how='any', axis=0)
        rho, p = br.model.spearman_r()
        file = spearman_file(name)
        rho.to_excel(file)

def run_uncertainty_and_sensitivity(name, N, rule='L',
                                    across_lines=False, 
                                    across_oil_content=False,
                                    derivative=False,
                                    sample_cache={},
                                    autosave=True,
                                    autoload=True,
                                    optimize=True,
                                    line=None,
                                    **kwargs):
    print(f"Running {name}!")
    filterwarnings('ignore', category=bst.exceptions.DesignWarning)
    filterwarnings('ignore', category=bst.exceptions.CostWarning)
    br = cane.Biorefinery(name, **kwargs)
    br.model.retry_evaluation = True
    file = monte_carlo_file(name, across_lines, across_oil_content, line)
    N_notify = min(int(N/10), 20)
    autosave = N_notify if autosave else False
    if across_lines:
        df = cane.get_composition_data()
        current_line = [line]
        if name == 'O2': 
            sugarcane_name = name.replace('O2', 'S2')
        else:
            sugarcane_name = name
        br_sugarcane = cane.Biorefinery(sugarcane_name, **kwargs, feedstock_line=line)
        def set_line(line, current_line=current_line):
            if name in ('O1', 'O2'): 
                config = br_sugarcane if line in ('WT', 'EC') else br
            else:
                config = br
            config.set_feedstock_line(line)
            current_line[0] = line
            np.random.seed(1)
            config.model.load_samples(config.model.sample(N, rule=rule), optimize=optimize)
        
        @no_derivative
        def evaluate(current_line=current_line, **kwargs):
            line = current_line[0]
            if name in ('O1', 'O2'): 
                config = br_sugarcane if line in ('WT', 'EC') else br
            else:
                config = br
            autoload_file = autoload_file_name(f"{name}_{config.feedstock_line}")
            config.model.evaluate(
                autosave=autosave, 
                autoload=autoload,
                file=autoload_file,
                **kwargs,
            )
            if config is br_sugarcane:
                br.model.table = br_sugarcane.model.table
            
        br.model.evaluate_across_coordinate(
            name='Line',
            notify=int(N/10),
            f_coordinate=set_line,
            f_evaluate=evaluate,
            coordinate=df.index,
            notify_coordinate=True,
            xlfile=file,
        )
    elif across_oil_content:
        evaluate = None
        parameter = br.set_cane_oil_content
        parameter.name = 'ROI target'
        parameter.element = 'Biorefinery'
        parameter.distribution = shape.Uniform(0.1, 0.2)
        br.set_ROI_target = parameter
        parameter.units = '%'
        del br.set_cane_oil_content
        np.random.seed(1)
        samples = br.model.sample(N, rule)
        
        def set_ROI_target(ROI_target):
            br.ROI_target = ROI_target
            
        br.set_ROI_target.setter = set_ROI_target
        coordinate = np.linspace(0, 0.1, 15)
        if across_oil_content == 'oilcane vs sugarcane':
            # Replace `set_cane_oil_content` parameter with `set_ROI_target`.
            # The actual distribution does not matter because these values are updated
            # on the first coordinate when the oil content is 0 (i.e., when the feedstock is sugarcane).
            if any([i in name for i in ('5', '6', '7', '8')]):
                br_sugarcane = None
                br.model.metrics = [br.ROI, br.competitive_biomass_yield, br.net_energy_production]
            else:
                # Only sugarcane needs the ROI metric (which gets added later)
                br.model.metrics = [br.competitive_biomass_yield, br.energy_competitive_biomass_yield]
                br_sugarcane = cane.Biorefinery(name.replace('O', 'S'))
        elif across_oil_content == 'microbial oil vs bioethanol':
            raise NotImplementedError('microbial oil vs bioethanol is not yet ready')
            # ethanol_name = name
            # for i, j in [('5', '1'), ('6', '2'), ('7', '1'), ('8', '2')]:
            #     ethanol_name = ethanol_name.replace(i, j)
            # br_ethanol = cane.Biorefinery(ethanol_name)
            # br.feedstock.price = br_ethanol.feedstock.price = 0.035 # Same feedstock, same price, but actual price does not matter
            # parameter = br_ethanol.set_cane_oil_content
            # parameter.setter = lambda obj: None # Dissable parameter
            # br_ethanol.model.metrics = [br_ethanol.ROI]
            # br.model.metrics = [br.competitive_microbial_oil_yield, br.energy_competitive_microbial_oil_yield] # Only interested in this
            # br.model.specification = lambda: None # No need to simulate before metric
            # br_ethanol.model.load_samples(samples)
            # br_sugarcane = cane.Biorefinery(ethanol_name.replace('O', 'S'))
        else:
            raise ValueError(
                "`across_oil_content` must be either 'oilcane vs sugarcane' "
               f"or 'microbial oil vs bioethanol', not {across_oil_content}"
            )
        
        if br_sugarcane:
            br_sugarcane.model.metrics = [br_sugarcane.ROI, br_sugarcane.net_energy_production]
            br_sugarcane.model.load_samples(samples)
        br.model.load_samples(samples, optimize=optimize)
        @no_derivative
        def evaluate(**kwargs):
            oil_content = int(1000 * br.composition_specification.oil)
            autoload_file = autoload_file_name(f"{name}_{oil_content}_oil_content")
            if across_oil_content == 'oilcane vs sugarcane':
                autoload_file += '_o_vs_s'
                if br.composition_specification.oil == 0.:
                    # Configurations 1 and 2 do not work at zero oil content, so gotta go with respective sugarcane configuration.
                    if br_sugarcane:
                        br_sugarcane.model.evaluate(
                            autosave=autosave, 
                            autoload=autoload,
                            file=autoload_file,
                            **kwargs,
                        )
                        br.model.table.iloc[:, :] = br_sugarcane.model.table
                        # Given the oil content, also find the oilcane biomass yield
                        # required to get the same ROI as sugarcane.
                        br.model.table[br.set_ROI_target.index] = column = br_sugarcane.model.table[br_sugarcane.ROI.index]
                        br.model._samples[:, br.model.parameters.index(br.set_ROI_target)] = column.values
                        br.model.table[br.competitive_biomass_yield.index] = br.baseline_dry_biomass_yield
                    else:
                        br.model.evaluate(
                            autosave=autosave, 
                            autoload=autoload,
                            file=autoload_file,
                            **kwargs,
                        )
                        br.model.table[br.set_ROI_target.index] = column = br.model.table[br.ROI.index]
                        br.model._samples[:, br.model.parameters.index(br.set_ROI_target)] = column.values
                        br.model.table[br.competitive_biomass_yield.index] = br.baseline_dry_biomass_yield
                else:
                    br.model.evaluate(
                        autosave=autosave, 
                        autoload=autoload,
                        file=autoload_file,
                        **kwargs,
                    )
            elif across_oil_content == 'microbial oil vs bioethanol':
                raise NotImplementedError('microbial oil vs bioethanol is not yet ready')
                # autoload_file += '_mo_vs_etoh'
                # assert name in ('O6', 'O5')
                # # For configurations 5 and 6, find the microbial oil yield required
                # # to get the same ROI as bioethanol production
                # if br.composition_specification.oil == 0.:
                #     br_sugarcane.model.evaluate(
                #         autosave=autosave, 
                #         autoload=autoload,
                #         file=autoload_file,
                #         **kwargs,
                #     )
                #     br.model.table[br.set_ROI_target.index] = column = br_sugarcane.model.table[br_sugarcane.ROI.index]
                # else:
                #     br_ethanol.composition_specification.load_oil_content(br.composition_specification.oil) # Load coordinate (oil content)
                #     br_ethanol.model.evaluate(
                #         autosave=autosave, 
                #         autoload=autoload,
                #         file=autoload_file,
                #         **kwargs,
                #     )
                #     br.model.table[br.set_ROI_target.index] = column = br_ethanol.model.table[br_ethanol.ROI.index]
                # br.model._samples[:, br.model.parameters.index(br.set_ROI_target)] = column.values
                # br.model.evaluate(
                #     autosave=autosave, 
                #     autoload=autoload,
                #     file=autoload_file,
                #     **kwargs,
                # )
            
        br.model.evaluate_across_coordinate(
            name='Oil content',
            notify=int(N/10),
            f_coordinate=br.composition_specification.load_oil_content,
            f_evaluate=evaluate,
            coordinate=coordinate,
            notify_coordinate=True,
            xlfile=file,
        )
    else:
        autoload_file = autoload_file_name(name, line)
        np.random.seed(1)
        samples = br.model.sample(N, rule)
        br.model.load_samples(samples, optimize=optimize)
        success = False
        if not derivative: br.disable_derivative()
        for i in range(3):
            try:
                if derivative and name not in ('O1', 'O2'): br.disable_derivative()
                br.model.evaluate(
                    notify=int(N/10),
                    autosave=autosave,
                    autoload=autoload,
                    file=autoload_file
                )
            except Exception as e:
                raise e from None
                warn('failed evaluation; restarting without cache')
            else:
                success = True
                if derivative: br.enable_derivative()
                break
        if not success:
            raise RuntimeError('evaluation failed')
        br.model.table.to_excel(file)
        br.model.table = br.model.table.dropna(how='all', axis=1)
        for i in br.model.metrics:
            if i.index not in br.model.table: br.model._metrics.remove(i)
        br.model.table = br.model.table.dropna(how='any', axis=0)
        rho, p = br.model.spearman_r(filter='omit nan')
        file = spearman_file(name, line)
        rho.to_excel(file)

run = run_uncertainty_and_sensitivity
    
def run_all(N, across_lines=False, rule='L', configurations=None,
            filter=None,**kwargs):
    if configurations is None: configurations = configuration_names
    for name in configurations:
        if filter and not filter(name): continue
        run_uncertainty_and_sensitivity(
            name, N, rule, across_lines, **kwargs
        )

def run_sugarcane_microbial_oil_and_ethanol(N=None):
    if N is None: N = 50
    filterwarnings('ignore')
    cane.YRCP2023()
    run_uncertainty_and_sensitivity('S1', N, line='WT')
    run_uncertainty_and_sensitivity('O7', N, line='WT')
    run_uncertainty_and_sensitivity('S2', N, line='WT')
    run_uncertainty_and_sensitivity('O8', N, line='WT')
    run_uncertainty_and_sensitivity('O1', N)
    run_uncertainty_and_sensitivity('O7', N)
    run_uncertainty_and_sensitivity('O2', N)
    run_uncertainty_and_sensitivity('O8', N)
    
def run_oilcane_microbial_oil_and_ethanol_across_oil_content(N=None):
    if N is None: N = 20
    filterwarnings('ignore')
    cane.YRCP2023()
    run_uncertainty_and_sensitivity('O7', N, across_oil_content='oilcane vs sugarcane')
    run_uncertainty_and_sensitivity('O8', N, across_oil_content='oilcane vs sugarcane')
    run_uncertainty_and_sensitivity('O1', N, across_oil_content='oilcane vs sugarcane')
    run_uncertainty_and_sensitivity('O2', N, across_oil_content='oilcane vs sugarcane')        
    
def run_oilcane_microbial_oil_and_ethanol_across_lines(N=None):
    if N is None: N = 50
    filterwarnings('ignore')
    cane.YRCP2023()
    run_uncertainty_and_sensitivity('O7', N, across_lines=True)
    run_uncertainty_and_sensitivity('O8', N, across_lines=True)
    run_uncertainty_and_sensitivity('O1', N, across_lines=True)
    run_uncertainty_and_sensitivity('O2', N, across_lines=True)        


# -*- coding: utf-8 -*-
"""
"""
from .chemicals import create_chemicals
import biosteam as bst
from biosteam import main_flowsheet as F
from .process_settings import BiorefinerySettings
from .systems import create_system
from .tea import create_tea

__all__ = ('Biorefinery',)

class Biorefinery:
    
    def __new__(cls, name=None, flowsheet_ID='corn',
                chemicals=None, biorefinery_settings=None, cache=None):
        flowsheet = bst.Flowsheet(flowsheet_ID)
        F.set_flowsheet(flowsheet)
        if name is None: name = 'conventional dry-grind'
        if cache and name in cache:
            return cache[name]
        elif name != 'conventional dry-grind':
            raise ValueError(f"'{name}' is not available; "
                              "only 'conventional dry-grind' is a valid name")
        self = super().__new__(cls)
        chems = chemicals or self.chemicals
        bst.settings.set_thermo(chems)
        self._biorefinery_settings = settings = biorefinery_settings or BiorefinerySettings()
        settings.load_process_settings()
        self.flowsheet = flowsheet
        sys = self.system
        sys.simulate()
        tea = self.TEA
        if 'corn' in flowsheet.ID:
            self.corn_sys = sys
            self.corn_tea = tea
        tea.IRR = tea.solve_IRR()
        self.all_areas = bst.process_tools.UnitGroup('All Areas', sys.units)
        self.__dict__.update(flowsheet.to_dict())
        return self
    
    
    @property
    def chemicals(self):
        try:
            chemicals = self._chemicals
        except:
            self._chemicals = chemicals = create_chemicals()
        return chemicals
    
    @property
    def biorefinery_settings(self):
        try:
            settings = self._biorefinery_settings
        except:
            self._biorefinery_settings = settings = BiorefinerySettings()
        return settings
    
    @property
    def system(self):
        flowsheet = getattr(self, 'flowsheet', None)
        try:
            system = self._system
        except:
            self._system = system = create_system(
                flowsheet=flowsheet,
                biorefinery_settings=self.biorefinery_settings)
        return system
    
    @property
    def TEA(self):
        system = self.system
        tea = system.TEA or create_tea(system, **self.biorefinery_settings.tea_parameters)
        return tea


# -*- coding: utf-8 -*-
"""
"""
from thermosteam.reaction import Reaction, ParallelReaction
import thermosteam as tmo
import biosteam as bst
Rxn = tmo.reaction.Reaction
ParallelRxn = tmo.reaction.ParallelReaction

__all__ = (
    'GlycerolysisReactor',
)

class GlycerolysisReactor(bst.CSTR):
    _ins_size_is_fixed = False
    _N_ins = 2
    _N_outs = 2
    T_default = 273.15 + 230
    P_default = 101325
    tau_default = 2
    
    def _setup(self):
        super()._setup()
        # End result is near 100% conversion of FFAs with selectivity
        # of 42% MAG, 47% DAG, 11% TAG:
        # [1] Erik Anderson. Superior Process Technologies. 
        # Glycerolysis for Lowering Free Fatty Acid Levels
        
        # [2] Kapil Mamtani, KavehShahbaz Mohammed M.Farid. 
        # Glycerolysis of free fatty acids: A review. 
        # https://doi.org/10.1016/j.rser.2020.110501
        
        self.glycerolysis_baseline = ParallelReaction([
          Reaction('FFA + Glycerol -> MAG + H2O', reactant='FFA',  X=1.00),
          Reaction('DAG + Glycerol -> 2MAG', reactant='DAG',  X=1.00),
          Reaction('TAG + 2Glycerol -> 3MAG', reactant='TAG',  X=1.00),
        ])
        self.glycerolysis = ParallelReaction([
          Reaction('2MAG -> DAG + Glycerol', reactant='MAG',  X=0.49),
          Reaction('3MAG -> TAG + 2Glycerol', reactant='MAG',  X=0.13),
        ])
        
        # Alternative preliminary modeling for legacy purposes:
        # self.glycerolysis = ParallelReaction([
        #   Reaction('FFA + Glycerol -> MAG + H2O', reactant='FFA',  X=0.80),
        #   Reaction('2FFA + Glycerol -> DAG + 2H2O', reactant='FFA',  X=0.15),
        #   Reaction('3FFA + Glycerol -> TAG + 3H2O', reactant='FFA',  X=0.05),
        # ])
        
    def _run(self):
        feed, *other, N2 = self.ins
        vent, effluent = self.outs
        vent.P = effluent.P = self.P
        vent.T = effluent.T = self.T
        vent.phase = 'g'
        effluent.mix_from(self.ins, energy_balance=False)
        self.glycerolysis_baseline.force_reaction(effluent)
        self.glycerolysis(effluent)
        effluent.mol.remove_negatives() # The correct glycerol flow rate is taken care of in a unit specification
        vent.copy_flow(effluent, ('N2', 'Water'), remove=True)
        


# -*- coding: utf-8 -*-
"""
Created on Fri Jun  4 23:44:10 2021

@author: yrc2
"""
from biorefineries import oilcane as oc
import biosteam as bst
import numpy as np
from biosteam.utils import colors, GG_colors
import matplotlib.pyplot as plt
from biosteam.plots import (
    plot_contour_2d,
    MetricBar,
    plot_vertical_line,
    rounded_tickmarks_from_data as tickmarks,
    plot_scatter_points,
)
from thermosteam.units_of_measure import format_units
from thermosteam.utils import set_figure_size, set_font
from biorefineries.oilcane._load_data import images_folder
from warnings import filterwarnings
from scipy.ndimage.filters import gaussian_filter
import os

__all__ = (
    'plot_relative_sorghum_oil_content_and_cane_oil_content_contours_manuscript',
    'plot_recovery_and_oil_content_contours_manuscript',
    'plot_recovery_and_oil_content_contours',
    'plot_relative_sorghum_oil_content_and_cane_oil_content_contours',
    'plot_recovery_and_oil_content_contours_biodiesel_only',
    'plot_recovery_and_oil_content_contours_with_oilsorghum_only',
    'plot_metrics_across_composition'
)

filterwarnings('ignore', category=bst.exceptions.DesignWarning)
line_colors = [
    GG_colors.orange.RGBn,
    GG_colors.purple.RGBn,
    GG_colors.green.RGBn,
    GG_colors.blue.RGBn,
    GG_colors.yellow.RGBn,
    colors.CABBI_teal.RGBn,
    colors.CABBI_grey.RGBn,
    colors.CABBI_brown.RGBn,
]
shadecolor = (*colors.neutral.RGBn, 0.20)
linecolor = (*colors.neutral_shade.RGBn, 0.85)
targetcolor = (*colors.red_tint.RGBn, 1)
startcolor = (*colors.red_tint.RGBn, 1)
edgecolor = (*colors.CABBI_black.RGBn, 1)
    
CABBI_colors = (colors.CABBI_yellow.tint(75).RGBn, 
                colors.CABBI_yellow.tint(30).RGBn,
                colors.CABBI_yellow.RGBn,
                colors.CABBI_green.tint(5).RGBn,
                colors.CABBI_teal_green.shade(40).RGBn,
                colors.CABBI_teal_green.shade(60).RGBn)

CABBI_colors_x = (colors.CABBI_blue_light.tint(90).RGBn,
                  colors.CABBI_blue_light.tint(40).RGBn, 
                  colors.CABBI_blue_light.RGBn, 
#                   colors.CABBI_teal.RGBn,
#                   colors.CABBI_teal_green.tint(10).RGBn,
                  colors.CABBI_teal_green.tint(40).shade(15).RGBn,
                  colors.CABBI_teal_green.shade(45).RGBn)

diverging_colormaps = [
    plt.cm.get_cmap('RdYlGn')
]

colormaps = [
    plt.cm.get_cmap('viridis'),
    plt.cm.get_cmap('copper_r'),
    # LinearSegmentedColormap.from_list('CABBI', CABBI_colors, 25),
    # LinearSegmentedColormap.from_list('CABBI', CABBI_colors_x, 25),
    plt.cm.get_cmap('inferno_r'),
    plt.cm.get_cmap('copper_r'),
    plt.cm.get_cmap('bone_r'),
] * 2

light_letter_color = colors.neutral.tint(98).RGBn
letter_color = colors.neutral.tint(80).RGBn
dark_letter_color = colors.neutral.shade(80).RGBn

# %% Plot functions for publication

def _add_letter_labels(axes, xpos, ypos, colors):
    M, N = shape = colors.shape
    if shape == (2, 2):
        letters=np.array([['A', 'C'], ['B', 'D']])
    elif shape == (3, 2):
        letters=np.array([['A', 'D'], ['B', 'E'], ['C', 'F']])
    elif shape == (2, 1):
        letters=np.array([['A'], ['B']])
    for i in range(M):
        for j in range(N):
            ax = axes[i, j]
            letter = letters[i, j]
            xlb, xub = ax.get_xlim()
            ylb, yub = ax.get_ylim()
            if hasattr(ax, '_cached_ytwin'):                
                ax = ax._cached_ytwin
            ax.text((xlb + xub) * xpos, (yub + ylb) * ypos, letter, color=colors[i, j],
                     horizontalalignment='center',verticalalignment='center',
                     fontsize=12, fontweight='bold', zorder=1e17)

def plot_metrics_across_composition_manuscript(load=True, fs=8, smooth=1):
    set_font(size=fs)
    set_figure_size()
    fig, axes = plot_metrics_across_composition(
        load=load, 
        smooth=smooth,
    )
    colors = np.zeros([2, 2], object)
    colors[:] = [[light_letter_color, light_letter_color],
                 [light_letter_color, light_letter_color],
                 [light_letter_color, light_letter_color]]
    _add_letter_labels(axes, 1 - 0.68, 0.7, colors)
    plt.subplots_adjust(right=0.92, wspace=0.1 * (fs/8) ** 2, top=0.9, bottom=0.10)
    for i in ('svg', 'png'):
        file = os.path.join(images_folder, f'recovery_and_oil_content_contours.{i}')
        plt.savefig(file, transparent=True)

def plot_recovery_and_oil_content_contours_manuscript(load=True, fs=8, smooth=1):
    set_font(size=fs)
    set_figure_size()
    fig, axes = plot_recovery_and_oil_content_contours(
        load=load, 
        smooth=smooth,
    )
    colors = np.zeros([2, 2], object)
    colors[:] = [[light_letter_color, light_letter_color],
                 [light_letter_color, light_letter_color]]
    _add_letter_labels(axes, 1 - 0.68, 0.7, colors)
    plt.subplots_adjust(right=0.92, wspace=0.1 * (fs/8) ** 2, top=0.9, bottom=0.10)
    for i in ('svg', 'png'):
        file = os.path.join(images_folder, f'recovery_and_oil_content_contours.{i}')
        plt.savefig(file, transparent=True)

def plot_recovery_and_oil_content_contours_with_oilsorghum_only(fs=10, smooth=1):
    set_font(size=fs)
    set_figure_size(4, 1.1)
    fig, axes = plot_recovery_and_oil_content_contours(
        load=True, 
        smooth=smooth,
        with_oilsorghum_only=True
    )
    colors = np.zeros([2, 1], object)
    colors[:] = [[light_letter_color], [light_letter_color]]
    # _add_letter_labels(axes, 1 - 0.68, 0.7, colors)
    plt.subplots_adjust(left=.2, right=0.92, wspace=0.1 * (fs/8) ** 2, top=0.9, bottom=0.10)
    for i in ('svg', 'png'):
        file = os.path.join(images_folder, f'recovery_and_oil_content_contours.{i}')
        plt.savefig(file, transparent=True)

def plot_recovery_and_oil_content_contours_biodiesel_only(load=True, fs=8, metric_indices=None):
    set_font(size=fs)
    set_figure_size()
    if metric_indices is None: metric_indices = (0, 2, 6, 10)
    for cmap, i in zip(colormaps, metric_indices):
        fig, axes = plot_recovery_and_oil_content_contours(
            load=load, configurations=np.array([[7, 8], [5, 6]]),
            N_points=20, yticks=[0, 2.5, 5, 7.5, 10, 12.5, 15],
            titles=['Batch', 'Fed-Batch'],
            metric_index=i, cmap=cmap,
        )
        load = True
        colors = np.zeros([2, 2], object)
        colors[:] = [[light_letter_color, light_letter_color],
                     [light_letter_color, light_letter_color]]
        _add_letter_labels(axes, 1 - 0.68, 0.85, colors)
        plt.subplots_adjust(right=0.92, wspace=0.1 * (fs/8) ** 2, top=0.9, bottom=0.10)
        for j in ('svg', 'png'):
            file = os.path.join(images_folder, f'recovery_and_oil_content_contours_biodiesel_only_{i}.{j}')
            plt.savefig(file, transparent=True)

def plot_relative_sorghum_oil_content_and_cane_oil_content_contours_manuscript(load=True, fs=8, smooth=0.9):
    set_font(size=fs)
    set_figure_size()
    fig, axes = plot_relative_sorghum_oil_content_and_cane_oil_content_contours(
        load=load, smooth=smooth,
    )
    colors = np.zeros([2, 2], object)
    colors[:] = [[light_letter_color, light_letter_color],
                 [light_letter_color, light_letter_color]]
    _add_letter_labels(axes, 1 - 0.82, 0.70, colors)
    plt.subplots_adjust(right=0.92, wspace=0.1, top=0.9, bottom=0.12)
    for i in ('svg', 'png'):
        file = os.path.join(images_folder, f'relative_sorghum_oil_content_and_cane_oil_content_contours.{i}')
        plt.savefig(file, transparent=True)

# %% General    

def metrics_across_oil_and_fiber_content(configuration, load):
    # Generate contour data
    x = np.linspace(0., 0.04, 5)
    y = np.linspace(0.35, 0.75, 5)
    z = np.array([0.6, 0.65, 0.7])
    X, Y, Z = np.meshgrid(x, y, z)
    folder = os.path.dirname(__file__)
    file = f'{configuration}_composition_analysis.npy'
    file = os.path.join(folder, file)
    if load:
        data = np.load(file, allow_pickle=True)
    else:
        from warnings import filterwarnings
        filterwarnings('ignore')
        # This returns data of all metrics for the given configuration,
        # but we are mainly interested in MFPP and productivity in L biodiesel per MT cane.
        data = oc.evaluate_metrics_across_composition(
            X, Y, Z, configuration,
        ) 
    np.save(file, data)
    return X, Y, Z, data

def plot_metrics_across_composition(
        configuration=None, load=False, N_decimals=1, 
        yticks=None, titles=None, 
        cmap=None, smooth=None,
    ):
    if configuration is None: configuration = 'O2'
    metric_indices=[0, 2]
    MFPP = oc.all_metric_mockups[0] # Maximum feedstock purchase price
    BP = oc.all_metric_mockups[2] # Biodiesel production
    # EP = oc.all_metric_mockups[5] # Energy production
    X, Y, Z, data = metrics_across_oil_and_fiber_content(load)
    data = data[:, :, :, metric_indices]
    xticks = [0,   1,   2,   3,   4]
    yticks = [35,  45,  55,  65,  75]
    if smooth: # Smooth curves due to heat exchanger network and discontinuities in design decisionss
        A, B, C, D = data.shape
        for i in range(C):
            for j in range(D):
                data[:, :, i, j] = gaussian_filter(data[:, :, i, j], smooth)
    data = np.swapaxes(data, 2, 3)
    # Plot contours
    xlabel = 'Oil content [dry wt. %]'
    ylabel = "Fiber content [dry wt. %]"
    if titles is None: titles = np.array(['60% moisture', '65% moisture', '70% moisture'])
    metric_bars = [
        MetricBar(MFPP.name, format_units(MFPP.units), colormaps[0], tickmarks(data[:, :, 0, :], 8, 1, expand=0, p=0.5), 10, 1),
        MetricBar('Biod. prod.', format_units(BP.units), plt.cm.get_cmap('copper'), tickmarks(data[:, :, 1, :], 8, 1, expand=0, p=0.5), 10, 1),
        # MetricBar(EP.name, format_units(EP.units), colormaps[2], tickmarks(data[:, :, 2, :], 5, 5, expand=0, p=5), 10, 1),
    ]
    fig, axes, CSs, CB = plot_contour_2d(
        100.*X[:, :, 0], 100.*Y[:, :, 0], titles, data, xlabel, ylabel, xticks, yticks, metric_bars, 
        styleaxiskw=dict(xtick0=True), label=True,
    )
    def determine_axis_column(moisture_content):
        for column, axis_moisture in enumerate([60, 65, 70]):
            if abs(moisture_content - axis_moisture) < 2.5:
                return column
        raise RuntimeError('could not determine axis with similar moisture content')
    
    df = oc.get_composition_data()
    names = df.index
    lines = []
    for name, color in zip(names, line_colors):
        data = df.loc[name]
        lines.append(
            (name, 
             data['Stem Oil (dw)']['Mean'] * 100, 
             data['Fiber (dw)']['Mean'] * 100,
             data['Water (wt)']['Mean'] * 100,
             color)
        )
    
    txtbox = dict(boxstyle='round', facecolor=colors.neutral.shade(20).RGBn, 
                  edgecolor='None', alpha=0.9)
    for *axes_columns, _ in axes:
        for (name, lipid, fiber, moisture, color) in lines:
            index = determine_axis_column(moisture)
            plt.sca(axes_columns[index]._cached_ytwin)
            plt.text(
                lipid + 0.1, fiber + 1, name, weight='bold', c=color,
                bbox=txtbox,
            )
            plot_scatter_points(
                [lipid], [fiber], marker='o', s=50, 
                color=color, edgecolor=edgecolor, clip_on=False, zorder=1e6,
            )
    return fig, axes

def relative_sorghum_oil_content_and_cane_oil_content_data(load, relative):
    # Generate contour data
    y = np.linspace(0.05, 0.15, 20)
    x = np.linspace(-0.03, 0., 20) if relative else np.linspace(0.02, 0.15, 20)
    X, Y = np.meshgrid(x, y)
    folder = os.path.dirname(__file__)
    file = 'oil_content_analysis.npy'
    if relative: file = 'relative_' + file
    file = os.path.join(folder, file)
    configurations = [1, 2]
    if load:
        data = np.load(file)
    else:
        data = oc.evaluate_configurations_across_sorghum_and_cane_oil_content(
            X, Y, configurations, relative,
        )
    np.save(file, data)
    return X, Y, data
    
def plot_relative_sorghum_oil_content_and_cane_oil_content_contours(
        load=False, configuration_index=..., relative=False, smooth=None,
    ):
    # Generate contour data
    X, Y, data = relative_sorghum_oil_content_and_cane_oil_content_data(load, relative)
    
    data = data[:, :, configuration_index, [0, 6]]
    
    # Plot contours
    xlabel = "Oil-sorghum oil content [dry wt. %]" 
    if relative: xlabel = ('relative ' + xlabel).capitalize()
    ylabel = 'Oilcane oil content\n[dry wt. %]'
    yticks = [5, 7.5, 10, 12.5, 15]
    xticks = [-3, -2, -1, 0] if relative else [2, 5, 7.5, 10, 12.5, 15]
    MFPP = oc.all_metric_mockups[0]
    TCI = oc.all_metric_mockups[6]
    if configuration_index == 0:
        Z = np.array(["Direct Cogeneration"])
        data = data[:, :, :, np.newaxis]
    elif configuration_index == 1:
        Z = np.array(["Integrated Co-Fermentation"])
        data = data[:, :, :, np.newaxis]
    elif configuration_index == ...:
        Z = np.array(["Direct Cogeneration", "Integrated Co-Fermentation"])
        data = np.swapaxes(data, 2, 3)
    else:
        raise ValueError('configuration index must be either 0 or 1')
    metric_bars = [
        [MetricBar(MFPP.name, format_units(MFPP.units), colormaps[0], tickmarks(data[:, :, 0, 0], 5, 1, expand=0, p=0.5), 10, 1),
         MetricBar(MFPP.name, format_units(MFPP.units), colormaps[0], tickmarks(data[:, :, 0, 1], 5, 1, expand=0, p=0.5), 10, 1)],
        [MetricBar(TCI.name, format_units(TCI.units), colormaps[1], tickmarks(data[:, :, 1, 0], 5, 5, expand=0, p=5), 10, 1),
         MetricBar(TCI.name, format_units(TCI.units), colormaps[1], tickmarks(data[:, :, 1, 1], 5, 5, expand=0, p=5), 10, 1)],
    ]
    
    if smooth: # Smooth curves due to heat exchanger network and discontinuities in design decisionss
        A, B, M, N = data.shape
        for m in range(M):
            for n in range(N):
                metric_data = data[:, :, m, n]
                data[:, :, m, n] = gaussian_filter(metric_data, smooth)
                # for a in range(A):
                #     values = metric_data[a, :]
                #     values.sort()
                #     p = np.arange(values.size)
                #     coeff = np.polyfit(p, values, 5)
                #     values[:] = np.polyval(coeff, p)
                # for b in range(B):
                #     values = metric_data[:, b]
                #     values.sort()
                #     p = np.arange(values.size)
                #     coeff = np.polyfit(p, values, 5)
                #     values[:] = np.polyval(coeff, p)
    
    fig, axes, CSs, CB = plot_contour_2d(
        100.*X, 100.*Y, Z, data, xlabel, ylabel, xticks, yticks, metric_bars, 
        styleaxiskw=dict(xtick0=True), label=True,
    )
    # for i in axes.flatten():
    #     plt.sca(i)
    #     plot_scatter_points([7], [10], marker='*', s=100, color=startcolor,
    #                         edgecolor=edgecolor, clip_on=False, zorder=3)
    return fig, axes
    
def plot_recovery_and_oil_content_contours(
        load=False, metric_index=0, N_decimals=1, configurations=None,
        N_points=20, yticks=None, titles=None, cmap=None, smooth=None,
        with_oilsorghum_only=False,
    ):
    if yticks is None: yticks = [5, 7.5, 10, 12.5, 15]
    if configurations is None:
        if with_oilsorghum_only:
            configurations = np.array([['O1*'], ['O2*']])
            if titles is None: titles = ['Oilcane & Oil-sorghum']
        else:
            configurations = np.array([['O1', 'O1*'], ['O2', 'O2*']])
            if titles is None: titles = ['Oilcane Only', 'Oilcane & Oil-sorghum']
        
    # Generate contour data
    x = np.linspace(0.40, 1.0, N_points)
    y = np.linspace(yticks[0] / 100, yticks[-1] / 100, N_points)
    X, Y = np.meshgrid(x, y)
    metric = bst.metric
    folder = os.path.dirname(__file__)
    file = "oil_extraction_analysis.npy"
    file = os.path.join(folder, file)
    
    if load:
        data = np.load(file)
    else:
        data = oc.evaluate_configurations_across_recovery_and_oil_content(
            X, Y, configurations,
        )
    np.save(file, data)
    data = data[:, :, :, :, metric_index]
    
    if with_oilsorghum_only:
        data = data[:, :, :, -1:]
    
    if smooth: # Smooth curves due to heat exchanger network and discontinuities in design decisionss
        A, B, M, N = data.shape
        for m in range(M):
            for n in range(N):
                metric_data = data[:, :, m, n]
                data[:, :, m, n] = gaussian_filter(metric_data, smooth)
    
    # Plot contours
    xlabel = 'Crushing mill oil recovery [%]'
    ylabel = "Oil content [dry wt. %]"
    ylabels = [f'Direct Cogeneration\n{ylabel}',
               f'Integrated Co-Fermentation\n{ylabel}']
    xticks = [40, 50, 60, 70, 80, 90, 100]
    
    metric = oc.all_metric_mockups[metric_index]
    units = metric.units if metric.units == '%' else format_units(metric.units)
    if cmap is None: cmap = colormaps[metric_index]
    mb = lambda x, name=None: MetricBar(
        metric.name if name is None else name, units if name is None else "", cmap, 
        tickmarks(data[:, :, x, :], 5, 0.1, expand=0, p=0.1, f=lambda x: round(x, N_decimals)),
        10, N_decimals=N_decimals
    )
    
    metric_bars = [mb(0), mb(1, "")]
    fig, axes, CSs, CB = plot_contour_2d(
        100.*X, 100.*Y, titles, data, xlabel, ylabels, xticks, yticks, metric_bars, 
        fillcolor=None, styleaxiskw=dict(xtick0=False), label=True,
    )
    M, N = configurations.shape
    for i in range(M):
        for j in range(N):
            ax = axes[i, j]
            plt.sca(ax)
            plt.fill_between([60, 90], [yticks[0]], [yticks[-1]], 
                              color=shadecolor,
                              linewidth=1)
            plot_vertical_line(60, ls='-.',
                                color=linecolor,
                                linewidth=1.0)
            plot_vertical_line(90, ls='-.',
                               color=linecolor,
                               linewidth=1.0)
            if hasattr(ax, '_cached_ytwin'):                
                plt.sca(ax._cached_ytwin)
            # plot_scatter_points([60], [10], marker='*', s=100, color=startcolor,
            #                     edgecolor=edgecolor, clip_on=False, zorder=3)
            # plot_scatter_points([ub], [15], marker='*', s=100, color=targetcolor,
            #                     edgecolor=edgecolor, clip_on=False, zorder=3)
    return fig, axes


# -*- coding: utf-8 -*-
"""
.. autofunction:: biorefineries.cane.systems.oilcane.create_oilcane_to_biodiesel_and_ethanol_1g
.. autofunction:: biorefineries.cane.systems.oilcane.create_oilcane_to_biodiesel_and_ethanol_combined_1_and_2g_post_fermentation_oil_separation
.. autofunction:: biorefineries.cane.systems.oilcane.create_oilcane_to_crude_oil_and_ethanol_1g
.. autofunction:: biorefineries.cane.systems.oilcane.create_oilcane_to_crude_oil_and_ethanol_combined_1_and_2g_post_fermentation_oil_separation

"""
from . import biodiesel_ethanol
from . import cellulosic_biodiesel_ethanol
from . import crude_oil_ethanol
from . import cellulosic_crude_oil_ethanol

__all__ = (

    *biodiesel_ethanol.__all__,
    *cellulosic_biodiesel_ethanol.__all__,
    *crude_oil_ethanol.__all__,
    *cellulosic_crude_oil_ethanol.__all__,
)

from .biodiesel_ethanol import *
from .cellulosic_biodiesel_ethanol import *
from .crude_oil_ethanol import *
from .cellulosic_crude_oil_ethanol import *




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Bioindustrial-Park: BioSTEAM's Premier Biorefinery Models and Results
# Copyright (C) 2020-, Yalin Li <mailto.yalin.li@gmail.com>,
#                      Sarang Bhagwat <sarangb2@illinois.edu>,
#                      Yoel Cortes-Pena <yoelcortes@gmail.com>
#
# This module is under the UIUC open-source license. See
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.

'''
References
----------
[1] Humbird et al., Process Design and Economics for Biochemical Conversion of
    Lignocellulosic Biomass to Ethanol: Dilute-Acid Pretreatment and Enzymatic
    Hydrolysis of Corn Stover; Technical Report NREL/TP-5100-47764;
    National Renewable Energy Lab (NREL), 2011.
    https://www.nrel.gov/docs/fy11osti/47764.pdf
[2] Davis et al., Process Design and Economics for the Conversion of Lignocellulosic
    Biomass to Hydrocarbon Fuels and Coproducts: 2018 Biochemical Design Case Update;
    NREL/TP-5100-71949; National Renewable Energy Lab (NREL), 2018.
    https://doi.org/10.2172/1483234

'''

# %%


import thermosteam as tmo
from thermosteam import Chemical, Chemicals
from thermosteam.functional import rho_to_V

__all__ = ('chemical_groups', 'create_chemicals', 'get_chemical_properties',)

_cal2joule = 4.184 # auom('cal').conversion_factor('J')

chemical_groups = dict(
    OtherSugars = ('Arabinose', 'Mannose', 'Galactose', 'Cellobiose', 'Sucrose'),
    SugarOligomers = ('GlucoseOligomer', 'XyloseOligomer', 'GalactoseOligomer',
                      'ArabinoseOligomer', 'MannoseOligomer'),
    OrganicSolubleSolids = ('AmmoniumAcetate', 'SolubleLignin', 'Extractives', 'CSL'),
    InorganicSolubleSolids = ('AmmoniumSulfate', 'NaOH', 'HNO3', 'NaNO3',
                              'BoilerChems', 'Na2SO4', 'NH4OH', 'CalciumLactate',
                              'CalciumAcetate', 'CalciumSuccinate'),
    Furfurals = ('Furfural', 'HMF'),
    OtherOrganics = ('Xylitol', 'LacticAcid', 'SuccinicAcid', 'EthylLactate',
                     'EthylAcetate', 'EthylSuccinate'),
    COSOxNOxH2S = ('NO', 'NO2', 'SO2', 'CO', 'H2S'),
    Proteins = ('Protein', 'Enzyme'),
    CellMass = ('WWTsludge', 'FermMicrobe'),
    # Theoretically P4O10 should be soluble, but it's the product of the
    # auto-populated combustion reactions so should in solid phase, however no
    # P4O10 will be generated in the system as no P-containing chemicals
    # are included in "combustibles"
    OtherInsolubleSolids = ('Tar', 'Ash', 'CalciumDihydroxide', 'CaSO4', 'P4O10',
                            'BaghouseBag', 'CoolingTowerChems', 'Polymer'),
    OtherStructuralCarbohydrates = ('Glucan', 'Xylan', 'Arabinan', 'Mannan',
                                    'Galactan'),
    SeparatelyListedOrganics = ('Ethanol', 'Glucose', 'Xylose', 'AceticAcid',
                                'Lignin', 'Acetate'),
    SpearatedlyListedOthers = ('H2O', 'NH3', 'H2SO4', 'CO2', 'CH4', 'O2', 'N2'),
    )

sugars = ('Glucose', 'Xylose', *chemical_groups['OtherSugars'],
          *chemical_groups['SugarOligomers'])

soluble_groups = ('OtherSugars', 'SugarOligomers', 'OrganicSolubleSolids',
                  'Furfurals', 'OtherOrganics', 'SeparatelyListedOrganics')
soluble_organics = sum([chemical_groups[i] for i in soluble_groups], ())

solubles = (*soluble_organics, *chemical_groups['InorganicSolubleSolids'], 'H2SO4')

insoluble_groups = ('Proteins', 'CellMass', 'OtherInsolubleSolids',
                    'OtherStructuralCarbohydrates')
insolubles = sum([chemical_groups[i] for i in insoluble_groups], ('Lignin', 'Acetate'))


COD_chemicals = (*soluble_organics, *chemical_groups['OtherStructuralCarbohydrates'],
                *chemical_groups['CellMass'],  *chemical_groups['Proteins'])

combustibles = (*COD_chemicals, 'NH3', 'NH4OH', 'NO', 'CO', 'H2S', 'CH4')

# Chemicals that will be modeled in Distillation/Flash units,
# list is in ascending order of Tb
# Xylitol is not included due to high Tm and Tb thus will stay in liquid phase
vle_chemicals = ('Ethanol', 'H2O', 'EthylAcetate', 'AceticAcid', 'EthylLactate',
                 'Furfural', 'EthylSuccinate', 'SuccinicAcid', 'LacticAcid', 'HMF')


# %%

def create_chemicals(set_thermo=True):
    chems = Chemicals([])
    def add_chemical(ID, ref=None, **data):
        chemical = Chemical(ID, **data) if ref is None else ref.copy(ID, **data)
        chems.append(chemical)
        return chemical

    H2O = add_chemical('H2O')

    ##### Gases #####
    add_chemical('O2', phase='g', Hf=0)
    add_chemical('N2', phase='g', Hf=0)
    add_chemical('CH4', phase='g')
    add_chemical('CO', search_ID='CarbonMonoxide', phase='g', Hf=-26400*_cal2joule)
    add_chemical('CO2', phase='g')
    add_chemical('NH3', phase='g', Hf=-10963*_cal2joule)
    add_chemical('NO', search_ID='NitricOxide', phase='g')
    add_chemical('NO2', phase='g')
    add_chemical('H2S', phase='g', Hf=-4927*_cal2joule)
    add_chemical('SO2', phase='g')

    ##### Soluble inorganics #####
    add_chemical('H2SO4', phase='l')
    add_chemical('HNO3', phase='l', Hf=-41406*_cal2joule)
    add_chemical('NaOH', phase='l')
    # Arggone National Lab active thermochemical tables, accessed 04/07/2020
    # https://atct.anl.gov/Thermochemical%20Data/version%201.118/species/?species_number=928
    add_chemical('NH4OH', search_ID='AmmoniumHydroxide', phase='l', Hf=-336719)
    add_chemical('CalciumDihydroxide', phase='s', Hf=-235522*_cal2joule)
    add_chemical('AmmoniumSulfate', phase='l', Hf=-288994*_cal2joule)
    add_chemical('NaNO3', phase='l', Hf=-118756*_cal2joule)
    # NIST https://webbook.nist.gov/cgi/cbook.cgi?ID=C7757826&Mask=2, accessed 04/07/2020
    add_chemical('Na2SO4', phase='l', Hf=-1356380)
    CaSO4 = add_chemical('CaSO4', phase='s', Hf=-342531*_cal2joule)
    # The default Perry 151 value is likely to be wrong, use another model instead
    CaSO4.Cn.move_up_model_priority('LASTOVKA_S', 0)

    ##### Soluble organics #####
    add_chemical('Ethanol')
    add_chemical('AceticAcid')
    Glucose = add_chemical('Glucose')
    # This one is more consistent with others
    Glucose.Cn.l.move_up_model_priority('DADGOSTAR_SHAW', 0)
    GlucoseOligomer = add_chemical('GlucoseOligomer', search_db=False, phase='l',
                                   formula='C6H10O5', Hf=-233200*_cal2joule)
    GlucoseOligomer.copy_models_from(Glucose, ['Hvap', 'Psat', 'Cn', 'mu', 'kappa'])
    Extractives = add_chemical('Extractives', search_ID='GluconicAcid', phase='l')
    # Ref [2] modeled this as gluconic acid, but here copy all properties from glucose
    Extractives.copy_models_from(Glucose)

    Xylose = add_chemical('Xylose')
    Xylose.copy_models_from(Glucose, ['Hvap', 'Psat', 'mu'])
    XyloseOligomer = add_chemical('XyloseOligomer', search_db=False, phase='l',
                                  formula='C5H8O4', Hf=-182100*_cal2joule)
    XyloseOligomer.copy_models_from(Xylose, ['Hvap', 'Psat', 'Cn', 'mu'])

    Sucrose = add_chemical('Sucrose', phase='l')
    Sucrose.Cn.move_up_model_priority('DADGOSTAR_SHAW', 0)
    add_chemical('Cellobiose', phase='l', Hf=-480900*_cal2joule)

    Mannose = add_chemical('Mannose', phase='l', Hf=Glucose.Hf)
    Mannose.copy_models_from(Glucose, ['Hvap', 'Psat', 'Cn', 'mu'])
    add_chemical('MannoseOligomer', GlucoseOligomer)

    Galactose = add_chemical('Galactose', phase='l', Hf=Glucose.Hf)
    Galactose.copy_models_from(Glucose, ['Hvap', 'Psat', 'Cn','mu'])
    add_chemical('GalactoseOligomer', GlucoseOligomer)

    Arabinose = add_chemical('Arabinose', phase='l', Hf=Xylose.Hf)
    Arabinose.copy_models_from(Xylose, ['Hvap', 'Psat', 'mu'])
    add_chemical('ArabinoseOligomer', XyloseOligomer)

    add_chemical('SolubleLignin', search_ID='Vanillin', phase='l', Hf=-108248*_cal2joule)
    Protein = add_chemical('Protein', search_db=False, phase='l',
                           formula='CH1.57O0.31N0.29S0.007', Hf=-17618*_cal2joule)
    Enzyme = add_chemical('Enzyme', search_db=False, phase='l',
                          formula='CH1.59O0.42N0.24S0.01', Hf=-17618*_cal2joule)

    # Properties of fermentation microbes copied from Z_mobilis as in ref [1]
    FermMicrobe = add_chemical('FermMicrobe', search_db=False, phase='l',
                               formula='CH1.8O0.5N0.2', Hf=-31169.39*_cal2joule)
    WWTsludge = add_chemical('WWTsludge', search_db=False, phase='s',
                             formula='CH1.64O0.39N0.23S0.0035', Hf=-23200.01*_cal2joule)

    Furfural = add_chemical('Furfural')
    # Tb from chemspider(chemenu database)
    # http://www.chemspider.com/Chemical-Structure.207215.html, accessed 04/07/2020
    # https://www.chemenu.com/products/CM196167, accessed 04/07/2020
    # Using Millipore Sigma's Pressure-Temperature Nomograph Interactive Tool at
    # https://www.sigmaaldrich.com/chemistry/solvents/learning-center/nomograph.html,
    # will give ~300°C at 760 mmHg if using the 115°C Tb at 1 mmHg (accessed 04/07/2020)
    # Hfus from NIST, accessed 04/24/2020
    # https://webbook.nist.gov/cgi/cbook.cgi?ID=C67470&Mask=4
    HMF = add_chemical('HMF', Hf=-99677*_cal2joule, Tb=291.5+273.15, Hfus=19800)
    HMF.copy_models_from(Furfural, ['V', 'Hvap', 'Psat', 'mu', 'kappa'])
    HMF.Dortmund.update(Furfural.Dortmund)

    # Hfus from NIST, condensed phase, accessed 04/07/2020
    # https://webbook.nist.gov/cgi/cbook.cgi?ID=C87990&Mask=4
    add_chemical('Xylitol', phase='l', Hf=-243145*_cal2joule, Hfus=-1118600)

    # Hfus from NIST, accessed 04/07/2020
    # https://webbook.nist.gov/cgi/cbook.cgi?ID=C50215&Mask=4
    LacticAcid = add_chemical('LacticAcid', Hfus=11340)

    SuccinicAcid = add_chemical('SuccinicAcid', phase_ref='s')
    # Density from chemspider, http://www.chemspider.com/Chemical-Structure.1078.html,
    # accessed 06/30/2020
    V = rho_to_V(1560, SuccinicAcid.MW)
    SuccinicAcid.V.s.add_model(V)
    # The default EQ105 values are off
    SuccinicAcid.V.l.move_up_model_priority('YEN_WOODS_SAT')

    add_chemical('EthylAcetate')
    # Hf from DIPPR value in Table 3 of Vatani et al., Int J Mol Sci 2007, 8 (5), 407–432
    add_chemical('EthylLactate', Hf=-695080)
    add_chemical('EthylSuccinate')

    ##### Soluble organic salts #####
    add_chemical('Acetate', phase='l', Hf=-108992*_cal2joule)
    add_chemical('AmmoniumAcetate', phase='l', Hf=-154701*_cal2joule)

    # Hf from a Ph.D. dissertation (Lactic Acid Production from Agribusiness Waste Starch
    # Fermentation with Lactobacillus Amylophilus and Its Cradle-To-Gate Life
    # Cycle Assessment as A Precursor to Poly-L-Lactide, by Andréanne Harbec)
    # The dissertation cited Cable, P., & Sitnai, O. (1971). The Manufacture of
    # Lactic Acid by the Fermentation of Whey: a Design and Cost Study.
    # Commonwealth Scientific and Industrial Research Organization, Australia,
    # which was also cited by other studies, but the origianl source cannot be found online
    CalciumLactate = add_chemical('CalciumLactate', phase='l', Hf=-1686100)
    # Hf from Lange's Handbook of Chemistry, 15th edn., Table 6.3, PDF page 631
    add_chemical('CalciumAcetate', phase='l', Hf=-1514730)

    # Solubility of CalciumSuccinate is 3.2 g/L in water as Ca2+ based on
    # Burgess and Drasdo, Polyhedron 1993, 12 (24), 2905–2911, which is 12.5 g/L as CaSA
    # Baseline CalciumSuccinate is ~14 g/L in fermentation broth, thus assumes all
    # CalciumSuccinate in liquid phase
    CalciumSuccinate = add_chemical('CalciumSuccinate', phase='l')
    # Cannot find data on Hf of CalciumSuccinate, estimate here assuming
    # Hrxn for Ca(OH)2 and SA and Ca(OH)2 and LA are the same
    CalciumSuccinate.Hf = CalciumLactate.Hf + (SuccinicAcid.Hf-2*LacticAcid.Hf)

    ##### Insoluble organics #####
    Glucan = add_chemical('Glucan', search_db=False, phase='s', formula='C6H10O5', Hf=-233200*_cal2joule)
    Glucan.copy_models_from(Glucose, ['Cn'])
    add_chemical('Mannan', Glucan)
    add_chemical('Galactan', Glucan)

    Xylan = add_chemical('Xylan', search_db=False, phase='s', formula='C5H8O4', Hf=-182100*_cal2joule)
    Xylan.copy_models_from(Xylose, ['Cn'])
    add_chemical('Arabinan', Xylan)

    add_chemical('Lignin', search_ID='Vanillin', phase='s', Hf=-108248*_cal2joule)

    ##### Insoluble organics #####
    # Holmes, Trans. Faraday Soc. 1962, 58 (0), 1916–1925, abstract
    # This is for auto-population of combustion reactions
    add_chemical('P4O10', phase='s', Hf=-713.2*_cal2joule)
    add_chemical('Ash', search_ID='CaO', phase='s', Hf=-151688*_cal2joule, HHV=0, LHV=0)
    # This is to copy the solid state of Xylose
    Tar = add_chemical('Tar', Xylose, phase_ref='s')
    Glucose.at_state('l')
    Xylose.at_state('l')
    Tar.at_state('s')

    ##### Mixtures #####
    # CSL is modeled as 50% water, 25% protein, and 25% lactic acid in ref [1]
    # did not model separately as only one price is given
    CSL = add_chemical('CSL', search_db=False, phase='l',
                       formula='CH2.8925O1.3275N0.0725S0.00175',
                       Hf=Protein.Hf/4+H2O.Hf/2+LacticAcid.Hf/4)

    # Boiler chemicals includes amine, ammonia, and phosphate,
    # did not model separately as composition unavailable and only one price is given
    add_chemical('BoilerChems', search_ID='DiammoniumPhosphate', phase='l')

    ##### Fillers #####
    Polymer = add_chemical('Polymer', search_db=False, phase='s', MW=1, Hf=0, HHV=0, LHV=0)
    Polymer.Cn.add_model(evaluate=0, name='Constant')
    add_chemical('BaghouseBag', Polymer)
    add_chemical('CoolingTowerChems', Polymer)

    ##### Handling missing properties #####
    # Heat capacity based on Cp of biomass (1.25 J/g/K)
    # from Leow et al., Green Chemistry 2015, 17 (6), 3584–3599
    for chemical in (CSL, Protein, Enzyme, WWTsludge, FermMicrobe):
        chemical.Cn.add_model(1.25*chemical.MW)

    # Molar volume following assumptions in the lipidcane biorefinery,
    # assume densities for solubles and insolubles to be 1e5 and 1540 kg/m3, respectively
    for chemical in chems:
        if chemical in vle_chemicals or chemical.locked_state=='g':
            continue
        V_l = tmo.functional.rho_to_V(1e5, chemical.MW)
        V_s = tmo.functional.rho_to_V(1540, chemical.MW)
        if chemical.locked_state == 'l':
            chemical.V.add_model(V_l, top_priority=True)
        elif chemical.locked_state == 's':
            chemical.V.add_model(V_s, top_priority=True)

    # The Lakshmi Prasad model gives negative kappa values for some chemicals
    for chemical in chems:
        if chemical.locked_state:
            try: chemical.kappa.move_up_model_priority('Lakshmi Prasad', -1)
            except: pass

    # Default missing properties of chemicals to those of water
    for chemical in chems: chemical.default()

    ##### Aliases and groups #####
    chems.compile()
    chems.set_alias('AmmoniumSulfate', 'NH4SO4')
    chems.set_alias('AmmoniumSulfate', '(NH4)2SO4')
    chems.set_alias('CalciumDihydroxide', 'Lime')
    chems.set_alias('CaSO4', 'Gypsum')
    chems.set_alias('Extractives', 'Extract')
    chems.set_alias('H2O', 'Water')
    chems.set_alias('H2SO4', 'SulfuricAcid')
    chems.set_alias('Na2SO4', 'SodiumSulfate')
    chems.set_alias('NH3', 'Ammonia')
    chems.set_alias('NH4OH', 'AmmoniumHydroxide')
    chems.define_group('sugars', sugars)
    chems.define_group('soluble_organics', soluble_organics)
    chems.define_group('solubles', solubles)
    chems.define_group('insolubles', insolubles)
    chems.define_group('cod', COD_chemicals)
    chems.define_group('combustibles', combustibles)
    chems.define_group('vle', vle_chemicals)

    if set_thermo: tmo.settings.set_thermo(chems)

    return chems


# %%

# =============================================================================
# Function to output chemical properties
# =============================================================================

def get_chemical_properties(chemicals, T, P, output=False):
    import pandas as pd
    formulas = [chemical.formula for chemical in chemicals]
    MWs = [chemical.MW for chemical in chemicals]
    Hfs = [chemical.Hf for chemical in chemicals]
    HHVs = [chemical.HHV for chemical in chemicals]
    LHVs = [chemical.LHV for chemical in chemicals]
    phases = []
    Tbs = []
    Psats = []
    Vs = []
    Cns = []
    mus = []
    kappas = []

    for chemical in chemicals:
        if chemical.locked_state:
            phases.append(chemical.phase_ref)
            Tbs.append('NA')
            try: Psats.append(chemical.Psat(T=T, P=P))
            except: Psats.append('')
            try: Vs.append(chemical.V(T=T, P=P))
            except: Vs.append('')
            try: Cns.append(chemical.Cn(T=T))
            except: Cns.append('')
            try: mus.append(chemical.mu(T=T, P=P))
            except: mus.append('')
            try: kappas.append(chemical.kappa(T=T, P=P))
            except: kappas.append('')
        else:
            ref_phase = chemical.get_phase(T=T, P=P)
            phases.append(f'variable, ref={ref_phase}')
            Tbs.append(chemical.Tb)
            try: Psats.append(chemical.Psat(T=T, P=P))
            except: Psats.append('')
            try: Vs.append(chemical.V(ref_phase, T=T, P=P))
            except: Vs.append('')
            try: Cns.append(chemical.Cn(ref_phase, T=T))
            except: Cns.append('')
            try: mus.append(chemical.mu(ref_phase, T=T, P=P))
            except: mus.append('')
            try: kappas.append(chemical.kappa(ref_phase, T=T, P=P))
            except: kappas.append('')

    properties = pd.DataFrame(
        {'ID': chemicals.IDs,
          'formula': formulas,
          'MW': MWs,
          'HHV': HHVs,
          'LHV': LHVs,
          'Hf': Hfs,
          'phase': phases,
          'boiling point': Tbs,
          'Psat': Psats,
          'V': Vs,
          'Cn': Cns,
          'mu': mus,
          'kappa': kappas}
        )

    if output:
        properties.to_excel('chemical_properties.xlsx', sheet_name='properties')

# get_chemical_properties(chems, 400, 101325, output=True)


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
import flexsolve as flx
import numpy as np
import biosteam as bst
from biosteam import SystemFactory
from .. import streams as s

from biorefineries.biodiesel import (
    create_lipid_wash_system,
)
__all__ = (
    'create_lipid_exctraction_system',
    'create_post_fermentation_oil_separation_system',
)

@SystemFactory(
    ID='lipid_exctraction_sys',
    ins=[s.fermentation_effluent],
    outs=[s.lipid, s.cellmass, s.wastewater],
)
def create_lipid_exctraction_system(ins, outs):
    fermentation_effluent, = ins
    lipid, cellmass, wastewater, = outs
    U401 = bst.SolidsCentrifuge('U401', fermentation_effluent, ['', ''],
        split=dict(cellmass=0.98, lipid=0.98),
        moisture_content=0.5,
        solids=['cellmass'],
    )
    U402 = bst.DrumDryer('U402', 
        (U401-0, 'dryer_air', 'dryer_natural_gas'), 
        ('', 'dryer_outlet_air', 'dryer_emissions'),
        moisture_content=0.18, split=0.,
        utility_agent='Steam',
    )
    # X401 = bst.ThermalOxidizer('X401', (U403-1, 'oxidizer_air'), 'oxidizer_emissions')
    U403 = bst.ScrewPress('U403', U402-0, split=dict(cellmass=1, lipid=0.3, Water=0.8),)
    bst.ConveyingBelt('U405', U403-0, cellmass)
    lipid_wash_sys = create_lipid_wash_system(ins=U403-1, outs=lipid, mockup=True)
    washed_lipid, spent_wash_water = lipid_wash_sys.outs
    bst.Mixer(ins=[spent_wash_water, U401-1], outs=wastewater)

@SystemFactory(
    ID='post_fermentation_oil_separation_sys',
    ins=[s.stillage],
    outs=[s.lipid, s.wastewater, s.evaporator_condensate], 
)
def create_post_fermentation_oil_separation_system(ins, outs, wastewater_concentration=None,
                                                   target_oil_and_solids_content=60, 
                                                   separate_cellmass=False):
    lipid, wastewater, evaporator_condensate = outs
    if separate_cellmass:     
        cellmass = bst.Stream('cellmass')
        outs.insert(1, cellmass)
    V605 = bst.MixTank('V605', ins)
    P606 = bst.Pump('P606', V605-0)
    EvX = bst.MultiEffectEvaporator('Ev607',
        ins=P606-0,
        P=(101325, 69682, 47057, 30953),
        V=0.90, V_definition='First-effect',
        thermo=lipid.thermo.ideal(),
        flash=False,
    )
    EvX.target_oil_and_solids_content = target_oil_and_solids_content # kg / m3
    EvX.remove_evaporators = False
    P_original = tuple(EvX.P)
    Pstart = P_original[0]
    Plast = P_original[-1]
    N = len(P_original)
    def x_oil(V): # Objective function for specification
        EvX.V = V
        EvX.run()
        effluent = EvX.outs[0]
        moisture = effluent.imass['Water']
        total = effluent.F_mass
        return EvX.target_oil_and_solids_content - 1000. * (1. - moisture / total)
    
    @EvX.add_specification(run=False)
    def adjust_evaporation():
        V_last = EvX.V
        x0 = 0.
        x1 = 0.5
        EvX.P = P_original
        EvX._reload_components = True
        y0 = x_oil(x0)
        if y0 <= 0.:
            EvX.V = x0
            return
        elif EvX.remove_evaporators:
            EvX._load_components()
            for i in range(1, N):
                if x_oil(1e-6) < 0.:
                    EvX.P = np.linspace(Pstart, Plast, N - 1)
                    EvX._reload_components = True
                else:
                    break    
            y1 = x_oil(x1)
            EvX.V = flx.IQ_interpolation(x_oil, x0, x1, y0, y1, x=V_last, ytol=1e-5, xtol=1e-6)
        elif x_oil(1e-6) < 0.:
            EvX.V = 1e-6
        else:
            y1 = x_oil(x1)
            EvX.V = flx.IQ_interpolation(x_oil, 1e-6, x1, y0, y1, x=V_last, ytol=1e-5, xtol=1e-6)
        
    P607 = bst.Pump('P607', EvX-0, P=101325.)
    C603_2 = bst.LiquidsSplitCentrifuge('C603_2', P607-0, (lipid, ''), 
                                        split={'Oil': 0.99,
                                               'Water': 0.0001})
    if separate_cellmass:        
        C603_3 = bst.SolidsCentrifuge('C603_3', C603_2-1, (cellmass, ''), 
                                      split={'Cellmass': 0.99}, solids=('Cellmass',))
        stream = C603_3-1
    else:
        stream = C603_2-1
    S601 = bst.Splitter('S601', ins=EvX-1, outs=['', evaporator_condensate], split=0.5)
    M601 = bst.Mixer('M601', [S601-0, stream], wastewater)
    M601.target_wastewater_concentration = 60. # kg / m3
    @M601.add_specification(run=True, impacted_units=[S601])
    def adjust_wastewater_concentration():
        concentrated_wastewater = M601.ins[1]
        waste = concentrated_wastewater.F_mass - concentrated_wastewater.imass['Water'] 
        current_concentration = waste / concentrated_wastewater.F_vol
        required_water = (1./M601.target_wastewater_concentration - 1./current_concentration) * waste * 1000.
        F_mass = S601.ins[0].F_mass
        if F_mass:
            split = required_water / F_mass
            if split < 0:
                split = 0.
            elif split > 1.:
                split = 1.
            S601.split[:] = split


# -*- coding: utf-8 -*-
"""
Created on Fri Nov  5 01:57:46 2021

@author: yrc2
"""
import os
import numpy as np
import pandas as pd
import biosteam as bst
from .biorefinery import Biorefinery
from biorefineries import cane
from biorefineries.tea.cellulosic_ethanol_tea import foc_table, capex_table
from thermosteam.utils import array_roundsigfigs

__all__ = (
    'save_detailed_expenditure_tables',
    'save_detailed_life_cycle_tables',   
)

def save_detailed_expenditure_tables(sigfigs=3):
    folder = os.path.dirname(__file__)
    folder = os.path.join(folder, 'results')
    filename = 'expenditures.xlsx'
    file = os.path.join(folder, filename)
    writer = pd.ExcelWriter(file)
    
    def get_sys(name):
        brf = Biorefinery(name)
        return brf.sys
    
    def get_tea(name):
        brf = Biorefinery(name)
        return brf.tea
    IDs = (
        'S1',
        'S2', 
        'O1', 
        'O2',
        'O7',
        'O8',
    )
    names = [
        'Sugarcane E-DC',
        'Sugarcane E-ICF',
        # 'Sugarcane B-DC',
        # 'Sugarcane B-ICF',
        'Oilcane EB-DC',
        'Oilcane EB-ICF',
        'Oilcane B-DC',
        'Oilcane B-ICF',
    ]
    syss = [get_sys(i) for i in IDs]
    teas = [get_tea(i) for i in IDs]
    product_IDs = ['advanced_ethanol', 'cellulosic_ethanol', 'biodiesel']
    tables = {
        'VOC': bst.report.voc_table(syss, product_IDs, names, with_products=True),
        'FOC': foc_table(teas, names),
        'CAPEX': capex_table(teas, names),
    }
    for key, table in tables.items(): 
        values = array_roundsigfigs(table.values, sigfigs=3, inplace=True)
        if key == 'CAPEX': # Bug in pandas
            for i, col in enumerate(table):
                table[col] = values[:, i]
        # if key == 'VOC':
        #     new_index = sorted(
        #         table.index, 
        #         key=lambda x: -abs(table.loc[x][1:].sum()) if x[0] == 'Raw materials' else 0,
        #     )
        #     tables[key] = table.reindex(new_index)
        table.to_excel(writer, key)
    writer.save()
    return tables
    
def save_detailed_life_cycle_tables(sigfigs=3, product=None):
    if product is None:
        product = 'biodiesel'
    try:
        # Energy allocation by gasoline gallon equivalent (GGE)
        bst.PowerUtility.define_property(
            name='energy', units='kW',
            fget=lambda power_utility: -power_utility.rate,
        )
        bst.Stream.define_property(
            name='energy', units='kJ/hr',
            # Ignore streams that are not ethanol (e.g. process water)
            fget=lambda stream: stream.LHV if stream.price else 0.,
        )
        
        # Economic/revenue allocation
        bst.PowerUtility.define_property(
            name='revenue', units='USD/hr',
            fget=lambda power_utility: -power_utility.cost,
        )
        bst.Stream.define_property(
            name='revenue', units='USD/hr',
            # Ignore streams that are not ethanol (e.g. ash disposal)
            fget=lambda stream: stream.cost if stream.price > 0. else 0.,
        )
    except:
        pass
    def get(configuration, name, RIN=False):
        brf = Biorefinery(configuration)
        return getattr(brf, name)

    GWP = cane.biorefinery.GWP
    bst.settings.define_impact_indicator(GWP, 'kg*CO2e')
    folder = os.path.dirname(__file__)
    folder = os.path.join(folder, 'results')
    filename = 'life_cycle.xlsx'
    file = os.path.join(folder, filename)
    writer = pd.ExcelWriter(file)
    index = ['Energy allocation',
             'Economic allocation',
             'Displacement allocation']
    if product == 'biodiesel':
        IDs = ('O1', 'O2', 'O7', 'O8')
        streams = [(get(i, 'cellulosic_based_diesel'), get(i, 'biomass_based_diesel')) for i in IDs]
        columns = [
            'Oilcane EB-DC [kg∙CO2e∙kg-1]',
            'Oilcane EB-ICF [kg∙CO2e∙kg-1]',
            'Oilcane B-DC [kg∙CO2e∙kg-1]',
            'Oilcane B-ICF [kg∙CO2e∙kg-1]',
        ]
    elif product == 'ethanol':
        IDs = ('S1', 'S2', 'O1', 'O2')
        streams = [(get(i, 'advanced_ethanol'), get(i, 'cellulosic_ethanol')) for i in IDs]
        columns = [
            'Sugarcane E-DC [kg∙CO2e∙kg-1]',
            'Sugarcane E-ICF [kg∙CO2e∙kg-1]',
            'Oilcane EB-DC [kg∙CO2e∙kg-1]',
            'Oilcane EB-ICF [kg∙CO2e∙kg-1]',
        ]
    else:
        raise ValueError(f"product '{product}' is not valid; valid options include 'biodiesel' and 'ethanol'")
    names = [i.split(' [')[0] for i in columns]
    methods = (f'GWP_{product}_allocation', f'GWP_{product}', f'GWP_{product}_displacement')
    syss = [get(i, 'sys') for i in IDs]
    values = np.zeros([len(methods), len(IDs)])
    for i, method in enumerate(methods):
        for j, name in enumerate(IDs):
            brf = Biorefinery(name)
            values[i, j] = getattr(brf, method)() 
    df_gwp = pd.DataFrame(values, index=index, columns=columns)
    tables = {
        'Inventory': bst.report.lca_inventory_table(
            syss, GWP, streams, names
        ),
        'Displacement allocation': bst.report.lca_displacement_allocation_table(
            syss, GWP, streams, product, names
        ),
        'Energy allocation factors': bst.report.lca_property_allocation_factor_table(
            syss, property='energy', units='GGE/hr', system_names=names, groups=('ethanol',),
        ),
        'Economic allocation factors': bst.report.lca_property_allocation_factor_table(
            syss, property='revenue', units='USD/hr', system_names=names, groups=('ethanol',),
        ),
        'Displacement allocation factors': bst.report.lca_displacement_allocation_factor_table(
            syss, streams, GWP, names, groups=(product,),
        ),
        f'GWP {product}': df_gwp,
    }
    for key, table in tables.items(): 
        array_roundsigfigs(table.values, sigfigs=3, inplace=True)
        table.to_excel(writer, key) 
    writer.save()
    return tables


# -*- coding: utf-8 -*-
"""
Created on Fri Nov  5 01:57:46 2021

@author: yrc2
"""
import os
import numpy as np
import pandas as pd
import biosteam as bst
import biorefineries.cornstover as cs
import biorefineries.oilcane as oc
from thermosteam.utils import array_roundsigfigs

__all__ = (
    'save_detailed_expenditure_tables',
    'save_detailed_life_cycle_tables',   
)

def save_detailed_expenditure_tables(sigfigs=3):
    folder = os.path.dirname(__file__)
    folder = os.path.join(folder, 'results')
    filename = 'expenditures.xlsx'
    file = os.path.join(folder, filename)
    writer = pd.ExcelWriter(file)
    
    def get_sys(name):
        oc.load(name)
        return oc.sys
    
    def get_tea(name):
        oc.load(name)
        return oc.tea
    IDs = (
        # 'S1',
        # 'S2', 
        'O1', 
        'O2'
    )
    names = [
        # 'Conventional Sugarcane Biorefinery',
        # 'Cellulosic Sugarcane Biorefinery',
        'Conventional Oilcane Biorefinery',
        'Cellulosic Oilcane Biorefinery',
    ]
    syss = [get_sys(i) for i in IDs]
    teas = [get_tea(i) for i in IDs]
    product_IDs = [oc.advanced_ethanol.ID, oc.cellulosic_ethanol.ID, oc.biodiesel.ID]
    tables = {
        'VOC': bst.report.voc_table(syss, product_IDs, names, with_products=True),
        'FOC': cs.foc_table(teas, names),
        'CAPEX': cs.capex_table(teas, names),
    }
    for key, table in tables.items(): 
        values = array_roundsigfigs(table.values, sigfigs=3, inplace=True)
        if key == 'CAPEX': # Bug in pandas
            for i, col in enumerate(table):
                table[col] = values[:, i]
        # if key == 'VOC':
        #     new_index = sorted(
        #         table.index, 
        #         key=lambda x: -abs(table.loc[x][1:].sum()) if x[0] == 'Raw materials' else 0,
        #     )
        #     tables[key] = table.reindex(new_index)
        table.to_excel(writer, key)
    writer.save()
    return tables
    
def save_detailed_life_cycle_tables(sigfigs=3):
    try:
        # Energy allocation by gasoline gallon equivalent (GGE)
        bst.PowerUtility.define_property(
            name='energy', units='kW',
            fget=lambda power_utility: -power_utility.rate,
        )
        bst.Stream.define_property(
            name='energy', units='kJ/hr',
            # Ignore streams that are not ethanol (e.g. process water)
            fget=lambda stream: stream.LHV if stream.price else 0.,
        )
        
        # Economic/revenue allocation
        bst.PowerUtility.define_property(
            name='revenue', units='USD/hr',
            fget=lambda power_utility: -power_utility.cost,
        )
        bst.Stream.define_property(
            name='revenue', units='USD/hr',
            # Ignore streams that are not ethanol (e.g. ash disposal)
            fget=lambda stream: stream.cost if stream.price > 0. else 0.,
        )
    except:
        pass
    def get(configuration, name, RIN=False):
        oc.load(configuration)
        return getattr(oc, name)
    bst.settings.define_impact_indicator(oc.GWP, 'kg*CO2e')
    folder = os.path.dirname(__file__)
    folder = os.path.join(folder, 'results')
    filename = 'life_cycle.xlsx'
    file = os.path.join(folder, filename)
    writer = pd.ExcelWriter(file)
    IDs = ('S1', 'S2', 'O1', 'O2')
    names = ['Conventional Sugarcane Biorefinery',
             'Cellulosic Sugarcane Biorefinery',
             'Conventional Oilcane Biorefinery',
             'Cellulosic Oilcane Biorefinery',]
    syss = [get(i, 'sys') for i in IDs]
    ethanol_streams = [(get(i, 'advanced_ethanol'), get(i, 'cellulosic_ethanol')) for i in IDs]
    index = ['Energy allocation',
             'Economic allocation',
             'Displacement allocation']
    columns = [
        'Sugarcane DC [kg∙CO2e∙kg-1]',
        'Sugarcane ICF [kg∙CO2e∙kg-1]',
        'Oilcane DC [kg∙CO2e∙kg-1]',
        'Oilcane ICF [kg∙CO2e∙kg-1]',
    ]
    methods = ('GWP_ethanol_allocation', 'GWP_ethanol', 'GWP_ethanol_displacement')
    values = np.zeros([len(methods), len(IDs)])
    for i, method in enumerate(methods):
        for j, name in enumerate(IDs):
            oc.load(name)
            values[i, j] = getattr(oc, method)() 
    df_gwp = pd.DataFrame(values, index=index, columns=columns)
    tables = {
        'Inventory': bst.report.lca_inventory_table(
            syss, oc.GWP, ethanol_streams, names
        ),
        'Displacement allocation': bst.report.lca_displacement_allocation_table(
            syss, oc.GWP, ethanol_streams, 'ethanol', names
        ),
        'Energy allocation factors': bst.report.lca_property_allocation_factor_table(
            syss, property='energy', units='GGE/hr', system_names=names, groups=('ethanol',),
        ),
        'Economic allocation factors': bst.report.lca_property_allocation_factor_table(
            syss, property='revenue', units='USD/hr', system_names=names, groups=('ethanol',),
        ),
        'Displacement allocation factors': bst.report.lca_displacement_allocation_factor_table(
            syss, ethanol_streams, oc.GWP, names, groups=('ethanol',),
        ),
        'GWP ethanol': df_gwp,
    }
    for key, table in tables.items(): 
        array_roundsigfigs(table.values, sigfigs=3, inplace=True)
        table.to_excel(writer, key) 
    writer.save()
    return tables